@@ -33,6 +33,7 @@
 #include "digest.h"
 #include "util.h"
 #include "uapi.h"
+#include "sd-boot.h"
 
 #define TPM_EVENT_LOG_MAX_ALGOS		64
 
@@ -449,6 +450,34 @@ __tpm_event_print(tpm_event_t *ev, tpm_event_bit_printer *print_fn)
 	hexdump(ev->event_data, ev->event_size, print_fn, 8);
 }
 
+void
+tpm_predicted_event_print(tpm_event_t *ev)
+{
+	__tpm_predicted_event_print(ev, (void (*)(const char *, ...)) printf);
+}
+
+void
+__tpm_predicted_event_print(tpm_event_t *ev, tpm_event_bit_printer *print_fn)
+{
+	print_fn("%05lx: event type=%s pcr=%d data=%u bytes\n",
+			ev->file_offset,
+			tpm_event_type_to_string(ev->event_type),
+			ev->pcr_index, ev->event_size);
+
+	if (ev->__parsed)
+		tpm_parsed_event_print(ev->__parsed, print_fn);
+
+	if (!ev->predicted_digest.algo) {
+		print_fn("  %-10s\n", "no predicted digest");
+	} else {
+		const tpm_evdigest_t *d = &ev->predicted_digest;
+		print_fn("  %-10s %s\n", d->algo->openssl_name, digest_print_value(d));
+	}
+
+	print_fn("  Data:\n");
+	hexdump(ev->event_data, ev->event_size, print_fn, 8);
+}
+
 static const tpm_evdigest_t *
 __tpm_event_rehash_efi_variable(const char *var_name, tpm_event_log_rehash_ctx_t *ctx)
 {
@@ -548,47 +577,6 @@ tpm_event_decode_uuid(const unsigned char *data)
 	return uuid;
 }
 
-/*
- * Handle IPL events, which grub2 and sd-boot uses to hide its stuff in
- */
-static void
-__tpm_event_grub_file_destroy(tpm_parsed_event_t *parsed)
-{
-	drop_string(&parsed->grub_file.device);
-	drop_string(&parsed->grub_file.path);
-}
-
-const char *
-__tpm_event_grub_file_describe(const tpm_parsed_event_t *parsed)
-{
-	static char buffer[1024];
-
-	if (parsed->grub_file.device == NULL)
-		snprintf(buffer, sizeof(buffer), "grub2 file load from %s", parsed->grub_file.path);
-	else
-		snprintf(buffer, sizeof(buffer), "grub2 file load from (%s)%s", parsed->grub_file.device, parsed->grub_file.path);
-	return buffer;
-}
-
-
-static const tpm_evdigest_t *
-__tpm_event_grub_file_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
-{
-	const struct grub_file_event *evspec = &parsed->grub_file;
-	const tpm_evdigest_t *md = NULL;
-
-	debug("  re-hashing %s\n", __tpm_event_grub_file_describe(parsed));
-	if (evspec->device == NULL || !strcmp(evspec->device, "crypto0")) {
-		debug("  assuming the file resides on system partition\n");
-		md = runtime_digest_rootfs_file(ctx->algo, evspec->path);
-	} else {
-		debug("  assuming the file resides on EFI boot partition\n");
-		md = runtime_digest_efi_file(ctx->algo, evspec->path);
-	}
-
-	return md;
-}
-
 /*
  * For files residing on the EFI partition, grub usually formats these as
  * (hdX,gptY)/EFI/BOOT/some.file
@@ -596,11 +584,11 @@ __tpm_event_grub_file_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *pa
  * omitted (eg for kernel and initrd).
  */
 static bool
-__tpm_event_grub_file_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, const char *value)
+__grub_file_parse(grub_file_t *grub_file, const char *value)
 {
 	if (value[0] == '/') {
-		parsed->grub_file.device = NULL;
-		parsed->grub_file.path = strdup(value);
+		grub_file->device = NULL;
+		grub_file->path = strdup(value);
 	} else if (value[0] == '(') {
 		char *copy = strdup(value);
 		char *path;
@@ -612,13 +600,97 @@ __tpm_event_grub_file_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, c
 
 		*path++ = '\0';
 
-		parsed->grub_file.device = strdup(copy + 1);
-		parsed->grub_file.path = strdup(path);
+		grub_file->device = strdup(copy + 1);
+		grub_file->path = strdup(path);
 		free(copy);
 	} else {
 		return false;
 	}
 
+	return true;
+}
+
+static const char *
+__grub_file_join(grub_file_t grub_file)
+{
+	static char path[PATH_MAX];
+
+	if (grub_file.device == NULL)
+		snprintf(path, sizeof(path), "%s", grub_file.path);
+	else
+		snprintf(path, sizeof(path), "(%s)%s", grub_file.device, grub_file.path);
+
+	return path;
+}
+
+static void
+__grub_file_destroy(grub_file_t *grub_file)
+{
+	drop_string(&grub_file->device);
+	drop_string(&grub_file->path);
+}
+
+/*
+ * Handle IPL events, which grub2 and sd-boot uses to hide its stuff in
+ */
+static void
+__tpm_event_grub_file_destroy(tpm_parsed_event_t *parsed)
+{
+	__grub_file_destroy(&parsed->grub_file);
+}
+
+const char *
+__tpm_event_grub_file_describe(const tpm_parsed_event_t *parsed)
+{
+	static char buffer[1024];
+
+	snprintf(buffer, sizeof(buffer), "grub2 file load from %s", __grub_file_join(parsed->grub_file));
+	return buffer;
+}
+
+
+static const tpm_evdigest_t *
+__tpm_event_grub_file_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
+{
+	const grub_file_event *evspec = &parsed->grub_file;
+	const tpm_evdigest_t *md = NULL;
+
+	debug("  re-hashing %s\n", __tpm_event_grub_file_describe(parsed));
+	if (evspec->device == NULL || !strcmp(evspec->device, "crypto0")
+	   || !strncmp(evspec->device, "cryptouuid", strlen("cryptouuid"))
+	   || !strncmp("/boot/", evspec->path, 6)) {
+		debug("  assuming the file resides on system partition\n");
+		md = runtime_digest_rootfs_file(ctx->algo, evspec->path);
+	} else {
+		if (sdb_is_boot_entry(evspec->path) && ctx->boot_entry_path) {
+			debug("  getting different boot entry file from EFI boot partition: %s\n",
+			      ctx->boot_entry_path);
+			md = runtime_digest_rootfs_file(ctx->algo, ctx->boot_entry_path);
+		} else
+		if (sdb_is_kernel(evspec->path) && ctx->boot_entry) {
+			debug("  getting different kernel from EFI boot partition: %s\n",
+			      ctx->boot_entry->image_path);
+			md = runtime_digest_efi_file(ctx->algo, ctx->boot_entry->image_path);
+		} else
+		if (sdb_is_initrd(evspec->path) && ctx->boot_entry) {
+			debug("  getting different initrd from EFI boot partition: %s\n",
+			      ctx->boot_entry->initrd_path);
+			md = runtime_digest_efi_file(ctx->algo, ctx->boot_entry->initrd_path);
+		} else {
+			debug("  assuming the file resides on EFI boot partition\n");
+			md = runtime_digest_efi_file(ctx->algo, evspec->path);
+		}
+	}
+
+	return md;
+}
+
+static bool
+__tpm_event_grub_file_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, const char *value)
+{
+	if (!__grub_file_parse(&parsed->grub_file, value))
+		return false;
+
 	parsed->event_subtype = GRUB_EVENT_FILE;
 	parsed->destroy = __tpm_event_grub_file_destroy;
 	parsed->rehash = __tpm_event_grub_file_rehash;
@@ -641,21 +713,87 @@ static const char *
 __tpm_event_grub_command_describe(const tpm_parsed_event_t *parsed)
 {
 	static char buffer[128];
+	static char *topic = NULL;
+
+	switch (parsed->event_subtype) {
+	case GRUB_EVENT_COMMAND:
+		topic = "grub2 command";
+		break;
+	case GRUB_EVENT_COMMAND_LINUX:
+		topic = "grub2 linux command";
+		break;
+	case GRUB_EVENT_COMMAND_INITRD:
+		topic = "grub2 initrd command";
+		break;
+	case GRUB_EVENT_KERNEL_CMDLINE:
+		topic = "grub2 kernel cmdline";
+		break;
+	}
+
+	snprintf(buffer, sizeof(buffer), "%s \"%s\"", topic, parsed->grub_command.string);
 
-	if (parsed->event_subtype == GRUB_EVENT_COMMAND)
-		snprintf(buffer, sizeof(buffer), "grub2 command \"%s\"", parsed->grub_command.string);
-	else
-		snprintf(buffer, sizeof(buffer), "grub2 kernel cmdline \"%s\"", parsed->grub_command.string);
 	return buffer;
 }
 
 static const tpm_evdigest_t *
 __tpm_event_grub_command_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
 {
-	if (parsed->grub_command.string == NULL)
-		return NULL;
+	char *str = NULL;
+	size_t sz = 0;
+	const tpm_evdigest_t *digest = NULL;
+	grub_file_t file;
 
-	return digest_compute(ctx->algo, parsed->grub_command.string, strlen(parsed->grub_command.string));
+	switch (parsed->event_subtype) {
+	case GRUB_EVENT_COMMAND:
+		str = strdup(parsed->grub_command.string);
+		break;
+	case GRUB_EVENT_COMMAND_LINUX:
+		if (ctx->boot_entry && parsed->grub_command.file.path) {
+			file = (grub_file_t) {
+				.device = parsed->grub_command.file.device,
+				.path = ctx->boot_entry->image_path,
+			};
+			sz = snprintf(NULL, 0, "linux %s %s", __grub_file_join(file), ctx->boot_entry->options);
+			str = malloc(sz + 1);
+			snprintf(str, sz + 1, "linux %s %s", __grub_file_join(file), ctx->boot_entry->options);
+			debug("Hashed linux command: %s\n", str);
+		} else
+			str = strdup(parsed->grub_command.string);
+		break;
+	case GRUB_EVENT_COMMAND_INITRD:
+		if (ctx->boot_entry && parsed->grub_command.file.path) {
+			file = (grub_file_t) {
+				.device = parsed->grub_command.file.device,
+				.path = ctx->boot_entry->initrd_path,
+			};
+			sz = snprintf(NULL, 0, "initrd %s", __grub_file_join(file));
+			str = malloc(sz + 1);
+			snprintf(str, sz + 1, "initrd %s", __grub_file_join(file));
+			debug("Hashed initrd command: %s\n", str);
+		} else
+			str = strdup(parsed->grub_command.string);
+		break;
+	case GRUB_EVENT_KERNEL_CMDLINE:
+		if (ctx->boot_entry && parsed->grub_command.file.path) {
+			file = (grub_file_t) {
+				.device = parsed->grub_command.file.device,
+				.path = ctx->boot_entry->image_path,
+			};
+			sz = snprintf(NULL, 0, "%s %s", __grub_file_join(file), ctx->boot_entry->options);
+			str = malloc(sz + 1);
+			snprintf(str, sz + 1, "%s %s", __grub_file_join(file), ctx->boot_entry->options);
+			debug("Hashed kernel cmdline: %s\n", str);
+		} else
+			str = strdup(parsed->grub_command.string);
+		break;
+	}
+
+	if (str) {
+		digest = digest_compute(ctx->algo, str, strlen(str));
+		free(str);
+	}
+
+	return digest;
 }
 
 /*
@@ -686,15 +824,29 @@ __tpm_event_grub_command_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed
 	keyword = copy;
 	arg = copy + wordlen;
 
+	if (!strcmp(keyword, "grub_cmd") && !strncmp(arg, "linux", strlen("linux"))) {
+		for (wordlen = 0; (cc = arg[wordlen]) && (cc != ' '); ++wordlen)
+			;
+		if (arg[wordlen] == ' ' && !__grub_file_parse(&parsed->grub_command.file, arg + wordlen + 1))
+			goto failed;
+		parsed->event_subtype = GRUB_EVENT_COMMAND_LINUX;
+	} else
+	if (!strcmp(keyword, "grub_cmd") && !strncmp(arg, "initrd", strlen("initrd"))) {
+		for (wordlen = 0; (cc = arg[wordlen]) && (cc != ' '); ++wordlen)
+			;
+		if (arg[wordlen] == ' ' && !__grub_file_parse(&parsed->grub_command.file, arg + wordlen + 1))
+			goto failed;
+		parsed->event_subtype = GRUB_EVENT_COMMAND_INITRD;
+	} else
 	if (!strcmp(keyword, "grub_cmd")) {
 		parsed->event_subtype = GRUB_EVENT_COMMAND;
 	} else
 	if (!strcmp(keyword, "kernel_cmdline")) {
+		if (!__grub_file_parse(&parsed->grub_command.file, arg))
+			goto failed;
 		parsed->event_subtype = GRUB_EVENT_KERNEL_CMDLINE;
-	} else {
-		free(copy);
-		return false;
-	}
+	} else
+		goto failed;
 
 	parsed->grub_command.string = strdup(arg);
 	for (argc = 0, s = strtok(arg, " \t"); s && argc < GRUB_COMMAND_ARGV_MAX - 1; s = strtok(NULL, " \t")) {
@@ -708,6 +860,10 @@ __tpm_event_grub_command_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed
 
 	free(copy);
 	return true;
+
+failed:
+	free(copy);
+	return false;
 }
 
 static void
@@ -790,8 +946,8 @@ static const tpm_evdigest_t *
 __tpm_event_systemd_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
 {
 	const uapi_boot_entry_t *boot_entry = ctx->boot_entry;
-	char initrd[2048];
-	char initrd_utf16[4096];
+	char cmdline[2048];
+	char cmdline_utf16[4096];
 	unsigned int len;
 
 	/* If no --next-kernel option was given, do not rehash anything */
@@ -804,15 +960,16 @@ __tpm_event_systemd_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *pars
 	}
 
 	debug("Next boot entry expected from: %s %s\n", boot_entry->title, boot_entry->version? : "");
-	snprintf(initrd, sizeof(initrd), "initrd=%s %s",
+	snprintf(cmdline, sizeof(cmdline), "initrd=%s %s",
 			path_unix2dos(boot_entry->initrd_path),
 			boot_entry->options? : "");
+	debug("Measuring Kernel command line: %s\n", cmdline);
 
-	len = (strlen(initrd) + 1) << 1;
-	assert(len <= sizeof(initrd_utf16));
-	__convert_to_utf16le(initrd, strlen(initrd) + 1, initrd_utf16, len);
+	len = (strlen(cmdline) + 1) << 1;
+	assert(len <= sizeof(cmdline_utf16));
+	__convert_to_utf16le(cmdline, strlen(cmdline) + 1, cmdline_utf16, len);
 
-	return digest_compute(ctx->algo, initrd_utf16, len);
+	return digest_compute(ctx->algo, cmdline_utf16, len);
 }
 
 /*
@@ -842,6 +999,19 @@ __tpm_event_tag_destroy(tpm_parsed_event_t *parsed)
 {
 }
 
+static const char *
+__tpm_event_tag_loader_conf_describe(const tpm_parsed_event_t *parsed)
+{
+	return "/loader/loader.conf (measured by systemd-boot)";
+}
+
+static const tpm_evdigest_t *
+__tpm_event_tag_loader_conf_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *parsed, tpm_event_log_rehash_ctx_t *ctx)
+{
+	debug("  re-hashing /loader/loader.conf");
+	return runtime_digest_efi_file(ctx->algo, "/loader/loader.conf");
+}
+
 static const char *
 __tpm_event_tag_options_describe(const tpm_parsed_event_t *parsed)
 {
@@ -876,10 +1046,12 @@ __tpm_event_tag_initrd_rehash(const tpm_event_t *ev, const tpm_parsed_event_t *p
 	}
 
 	debug("Next boot entry expected from: %s %s\n", boot_entry->title, boot_entry->version? : "");
+	debug("Measuring initrd: %s\n", boot_entry->initrd_path);
 	return runtime_digest_efi_file(ctx->algo, boot_entry->initrd_path);
 }
 
 /*
+ * Generated by systemd-boot (PCR#5), to measure loader.conf
  * Generated by the kernel (PCR#9), to measure the cmdline and initrd
  */
 static bool
@@ -900,6 +1072,10 @@ __tpm_event_parse_tag(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
 		return false;
 
 	parsed->destroy = __tpm_event_tag_destroy;
+	if (evspec->event_id == LOADER_CONF_EVENT_TAG_ID) {
+		parsed->rehash = __tpm_event_tag_loader_conf_rehash;
+		parsed->describe = __tpm_event_tag_loader_conf_describe;
+	} else
 	if (evspec->event_id == LOAD_OPTIONS_EVENT_TAG_ID) {
 		parsed->rehash = __tpm_event_tag_options_rehash;
 		parsed->describe = __tpm_event_tag_options_describe;
@@ -913,6 +1089,8 @@ __tpm_event_parse_tag(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
 	return true;
 }
 
+#define GRUB_PREP_ENVBLK "PReP ENV Block"
+
 static bool
 __tpm_event_parse_ipl(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
 {
@@ -933,8 +1111,12 @@ __tpm_event_parse_ipl(tpm_event_t *ev, tpm_parsed_event_t *parsed, buffer_t *bp)
 	if (ev->pcr_index == 8)
 		return __tpm_event_grub_command_event_parse(ev, parsed, value);
 
-	if (ev->pcr_index == 9)
-		return __tpm_event_grub_file_event_parse(ev, parsed, value);
+	if (ev->pcr_index == 9) {
+		if (strncmp(value, GRUB_PREP_ENVBLK, strlen(GRUB_PREP_ENVBLK)) == 0)
+			return __tpm_event_grub_envblk_event_parse(ev, parsed, value);
+		else
+			return __tpm_event_grub_file_event_parse(ev, parsed, value);
+	}
 
 	if (ev->pcr_index == 12)
 		return __tpm_event_systemd_event_parse(ev, parsed, value, len);
@@ -970,6 +1152,9 @@ __tpm_event_parse(tpm_event_t *ev, tpm_parsed_event_t *parsed, tpm_event_log_sca
 
 	case TPM2_EFI_GPT_EVENT:
 		return __tpm_event_parse_efi_gpt(ev, parsed, &buf);
+
+	case TPM2_EVENT_COMPACT_HASH:
+		return __tpm_event_parse_compact_hash(ev, parsed, &buf);
 	}
 
 	return false;
@@ -1001,6 +1186,7 @@ tpm_event_log_rehash_ctx_init(tpm_event_log_rehash_ctx_t *ctx, const tpm_algo_in
 void
 tpm_event_log_rehash_ctx_destroy(tpm_event_log_rehash_ctx_t *ctx)
 {
+	buffer_free(ctx->sbatlevel);
 }
 
 void
