@@ -21,6 +21,7 @@
 #ifndef EVENTLOG_H
 #define EVENTLOG_H
 
+#include "digest.h"
 #include "types.h"
 
 typedef struct tpm_event {
@@ -42,6 +43,8 @@ typedef struct tpm_event {
 
 	/* set by the predictor during pre-scan */
 	int			rehash_strategy;
+
+	tpm_evdigest_t		predicted_digest;
 } tpm_event_t;
 
 typedef void			tpm_event_bit_printer(const char *, ...);
@@ -89,15 +92,20 @@ enum {
 enum {
 	/* IPL subtypes for grub */
 	GRUB_EVENT_COMMAND		     = 0x0001,
-	GRUB_EVENT_FILE			     = 0x0002,
-	GRUB_EVENT_KERNEL_CMDLINE	     = 0x0003,
-	SHIM_EVENT_VARIABLE		     = 0x0004,
-	SYSTEMD_EVENT_VARIABLE		     = 0x0005,
+	GRUB_EVENT_COMMAND_LINUX	     = 0x0002,
+	GRUB_EVENT_COMMAND_INITRD	     = 0x0003,
+	GRUB_EVENT_FILE			     = 0x0004,
+	GRUB_EVENT_KERNEL_CMDLINE	     = 0x0005,
+	SHIM_EVENT_VARIABLE		     = 0x0006,
+	SYSTEMD_EVENT_VARIABLE		     = 0x0007,
 };
 
 enum {
-	LOAD_OPTIONS_EVENT_TAG_ID	     = 0x8F3B22EDU,
-	INITRD_EVENT_TAG_ID		     = 0x8F3B22ECU,
+	/* systemd-boot */
+	LOADER_CONF_EVENT_TAG_ID             = 0xF5BC582A,
+	/* kernel */
+	LOAD_OPTIONS_EVENT_TAG_ID	     = 0x8F3B22ED,
+	INITRD_EVENT_TAG_ID		     = 0x8F3B22EC,
 };
 
 #define EFI_DEVICE_PATH_MAX		16
@@ -202,11 +210,21 @@ typedef struct tpm_event_log_rehash_ctx {
 	const pecoff_image_info_t *next_stage_img;
 
 	/* This get set when the user specifies --next-kernel */
+	char *			boot_entry_path;
 	uapi_boot_entry_t *	boot_entry;
+
+	buffer_t *		sbatlevel;
 } tpm_event_log_rehash_ctx_t;
 
 #define GRUB_COMMAND_ARGV_MAX	32
 
+typedef struct grub_file {
+	char *			device;
+	char *			path;
+} grub_file_t;
+
+typedef grub_file_t		grub_file_event;
+
 /*
  * Parsed event types
  */
@@ -246,13 +264,12 @@ typedef struct tpm_parsed_event {
 		/* for GRUB_COMMAND, GRUB_KERNEL_CMDLINE */
 		struct grub_command_event {
 			char *		string;
-			char *		argv[GRUB_COMMAND_ARGV_MAX];
+			char *          argv[GRUB_COMMAND_ARGV_MAX];
+			grub_file_t	file;
 		} grub_command;
 
-		struct grub_file_event {
-			char *		device;
-			char *		path;
-		} grub_file;
+		/* for GRUB_FILE */
+		grub_file_event		grub_file;
 
 		struct shim_event {
 			char *		string;
@@ -260,7 +277,7 @@ typedef struct tpm_parsed_event {
 		} shim_event;
 
 		struct efi_gpt_event {
-			char *		efi_partition;
+			char *		sys_partition;
 			char *		disk_device;
 		} efi_gpt_event;
 
@@ -274,6 +291,14 @@ typedef struct tpm_parsed_event {
 			uint32_t	event_data_len;
 			char		event_data[52];
 		} tag_event;
+
+		struct compact_hash_event {
+			char *		prep_partition;
+		} compact_hash_event;
+
+		struct grub_envblk_event {
+			char *		prep_partition;
+		} grub_envblk_event;
 	};
 } tpm_parsed_event_t;
 
@@ -287,6 +312,8 @@ extern unsigned int		event_log_get_event_count(const tpm_event_log_reader_t *log
 extern unsigned int		event_log_get_tpm_version(const tpm_event_log_reader_t *log);
 extern void			tpm_event_print(tpm_event_t *ev);
 extern void			__tpm_event_print(tpm_event_t *ev, tpm_event_bit_printer *print_fn);
+extern void			tpm_predicted_event_print(tpm_event_t *ev);
+extern void			__tpm_predicted_event_print(tpm_event_t *ev, tpm_event_bit_printer *print_fn);
 extern void			tpm_event_log_rehash_ctx_init(tpm_event_log_rehash_ctx_t *,
 					const tpm_algo_info_t *);
 extern void			tpm_event_log_rehash_ctx_destroy(tpm_event_log_rehash_ctx_t *);
@@ -323,4 +350,8 @@ extern bool			shim_variable_name_valid(const char *name);
 extern const char *		shim_variable_get_rtname(const char *name);
 extern const char *		shim_variable_get_full_rtname(const char *name);
 
+extern bool			secure_boot_enabled();
+
+extern bool			__tpm_event_parse_compact_hash(tpm_event_t *, tpm_parsed_event_t *, buffer_t *);
+extern bool			__tpm_event_grub_envblk_event_parse(tpm_event_t *, tpm_parsed_event_t *, const char *);
 #endif /* EVENTLOG_H */
