@@ -23,6 +23,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <limits.h>
 
 #include "oracle.h"
 #include "util.h"
@@ -92,6 +93,7 @@ enum {
 	OPT_RSA_PUBLIC_KEY,
 	OPT_RSA_GENERATE_KEY,
 	OPT_RSA_BITS,
+	OPT_ECC_SRK,
 	OPT_INPUT,
 	OPT_OUTPUT,
 	OPT_AUTHORIZED_POLICY,
@@ -101,6 +103,7 @@ enum {
 	OPT_POLICY_FORMAT,
 	OPT_TARGET_PLATFORM,
 	OPT_BOOT_ENTRY,
+	OPT_COMPARE_CURRENT,
 };
 
 static struct option options[] = {
@@ -125,6 +128,7 @@ static struct option options[] = {
 	{ "public-key",		required_argument,	0,	OPT_RSA_PUBLIC_KEY },
 	{ "rsa-generate-key",	no_argument,		0,	OPT_RSA_GENERATE_KEY },
 	{ "rsa-bits",		required_argument,	0,	OPT_RSA_BITS },
+	{ "ecc-srk",		no_argument,		0,	OPT_ECC_SRK },
 	{ "input",		required_argument,	0,	OPT_INPUT },
 	{ "output",		required_argument,	0,	OPT_OUTPUT },
 	{ "authorized-policy",	required_argument,	0,	OPT_AUTHORIZED_POLICY },
@@ -134,6 +138,7 @@ static struct option options[] = {
 	{ "policy-format",	required_argument,	0,	OPT_POLICY_FORMAT },
 	{ "target-platform",	required_argument,	0,	OPT_TARGET_PLATFORM },
 	{ "next-kernel",	required_argument,	0,	OPT_BOOT_ENTRY },
+	{ "compare-current",	no_argument,		0,	OPT_COMPARE_CURRENT },
 
 	{ NULL }
 };
@@ -366,6 +371,7 @@ pcr_bank_extend_register(tpm_pcr_bank_t *bank, unsigned int pcr_index, const tpm
 static void
 predictor_extend_hash(struct predictor *pred, unsigned int pcr_index, const tpm_evdigest_t *d)
 {
+	debug("Extend PCR#%d: %s\n", pcr_index, digest_print(d));
 	pcr_bank_extend_register(&pred->prediction, pcr_index, d);
 }
 
@@ -481,7 +487,7 @@ __check_stop_event(tpm_event_t *ev, int type, const char *value, tpm_event_log_s
  * Lookahead: when processing the GPT event, we need to know which hard disk
  * we're talking about.
  */
-static void
+static bool
 __predictor_lookahead_efi_partition(tpm_event_t *ev, tpm_event_log_rehash_ctx_t *ctx)
 {
 	struct efi_gpt_event *gpt = &ev->__parsed->efi_gpt_event;
@@ -496,7 +502,31 @@ __predictor_lookahead_efi_partition(tpm_event_t *ev, tpm_event_log_rehash_ctx_t
 		if (!(parsed = ev->__parsed))
 			continue;
 
-		assign_string(&gpt->efi_partition, parsed->efi_bsa_event.efi_partition);
+		assign_string(&gpt->sys_partition, parsed->efi_bsa_event.efi_partition);
+		return true;
+	}
+	return false;
+}
+
+static void
+__predictor_lookahead_prep_partition(tpm_event_t *ev, tpm_event_log_rehash_ctx_t *ctx)
+{
+	struct efi_gpt_event *gpt = &ev->__parsed->efi_gpt_event;
+
+	while ((ev = ev->next) != NULL) {
+		tpm_parsed_event_t *parsed;
+
+		if (ev->event_type != TPM2_EVENT_COMPACT_HASH)
+			continue;
+
+		/* Compact hash events have already been parsed during the pre-scan */
+		if (!(parsed = ev->__parsed))
+			continue;
+
+		if (!parsed->compact_hash_event.prep_partition)
+			continue;
+
+		assign_string(&gpt->sys_partition, parsed->compact_hash_event.prep_partition);
 		return;
 	}
 }
@@ -579,6 +609,12 @@ predictor_get_event_strategy(unsigned int event_type)
 		 */
 		TPM2_EFI_GPT_EVENT,
 
+		/*
+		 * EVENT_COMPACT_HASH: used by SLOF for PCR4, to measure the bootloader with
+		 *	'BOOTLOADER' in the event data.
+		 */
+		TPM2_EVENT_COMPACT_HASH,
+
 		-1,
 	};
 	static int copy_types[] = {
@@ -628,6 +664,11 @@ predictor_pre_scan_eventlog(struct predictor *pred, tpm_event_t **stop_event_p)
 		ev->rehash_strategy = predictor_get_event_strategy(ev->event_type);
 		/* debug("%s -> %d\n", tpm_event_type_to_string(ev->event_type), ev->rehash_strategy); */
 
+		/* Calculate the values of PCR1 and PCR3 only based on the event data */
+		if (ev->rehash_strategy == EVENT_STRATEGY_PARSE_REHASH &&
+		    (ev->pcr_index == 1 || ev->pcr_index == 3))
+			ev->rehash_strategy = EVENT_STRATEGY_COPY;
+
 		if (ev->rehash_strategy == EVENT_STRATEGY_PARSE_REHASH) {
 			if (!tpm_event_parse(ev, &scan_ctx)) {
 				/* Provide better error logging */
@@ -652,6 +693,7 @@ predictor_update_eventlog(struct predictor *pred)
 	tpm_event_log_rehash_ctx_t rehash_ctx;
 	tpm_event_t *ev, *stop_event = NULL;
 	bool okay = true;
+	char boot_entry_path[PATH_MAX];
 
 	predictor_pre_scan_eventlog(pred, &stop_event);
 
@@ -662,9 +704,13 @@ predictor_update_eventlog(struct predictor *pred)
 	 * systemd ID of the next kernel entry to be booted.
 	 * FIXME: we should probably hide this behind a target_platform function.
 	 */
-	if (pred->boot_entry_id != NULL
-	 && !(rehash_ctx.boot_entry = sdb_identify_boot_entry(pred->boot_entry_id)))
-		fatal("unable to identify next kernel \"%s\"\n", pred->boot_entry_id);
+	if (pred->boot_entry_id != NULL) {
+		snprintf(boot_entry_path, sizeof(boot_entry_path),
+			 "%s/%s", UAPI_BOOT_DIRECTORY, pred->boot_entry_id);
+		assign_string(&rehash_ctx.boot_entry_path, boot_entry_path);
+		if (!(rehash_ctx.boot_entry = sdb_identify_boot_entry(pred->boot_entry_id)))
+			fatal("unable to identify next kernel \"%s\"\n", pred->boot_entry_id);
+	}
 
 	for (ev = pred->event_log; ev; ev = ev->next) {
 		tpm_evdigest_t *pcr;
@@ -705,9 +751,16 @@ predictor_update_eventlog(struct predictor *pred)
 			 * BOOT_SERVICES event that would tell us which partition we're booting
 			 * from.
 			 * Scan ahead to the first BSA event to extract the EFI partition.
+			 *
+			 * For PowerPC 64, the firmware measures the bootloader in a Compact
+			 * Hash event instead of a BSA event. If there is no BSA event, we
+			 * go further to look for the Compact Hash event to set the PReP
+			 * partition.
 			 */
-			if (ev->event_type == TPM2_EFI_GPT_EVENT)
-				__predictor_lookahead_efi_partition(ev, &rehash_ctx);
+			if (ev->event_type == TPM2_EFI_GPT_EVENT) {
+				if (!__predictor_lookahead_efi_partition(ev, &rehash_ctx))
+					__predictor_lookahead_prep_partition(ev, &rehash_ctx);
+			}
 
 			/* The shim loader emits an event that tells us which certificate it
 			 * used to verify the second stage loader. We try to predict that
@@ -758,6 +811,8 @@ predictor_update_eventlog(struct predictor *pred)
 			}
 
 			predictor_extend_hash(pred, ev->pcr_index, new_digest);
+
+			memcpy(&ev->predicted_digest, new_digest, sizeof(tpm_evdigest_t));
 		}
 
 no_action:
@@ -886,6 +941,87 @@ predictor_verify(struct predictor *pred, const char *source)
 	return num_mismatches;
 }
 
+static bool
+compare_events(struct predictor *pred, struct predictor *pred_cmp,
+	       unsigned int pcr_index, tpm_event_t *stop_event)
+{
+	tpm_event_t *ev, *ev_cmp;
+	const tpm_evdigest_t *predicted_digest, *cmp_digest;
+
+	for(ev = pred->event_log, ev_cmp = pred_cmp->event_log; ev; ev = ev->next, ev_cmp = ev_cmp->next) {
+		bool stop = false;
+		stop = (ev == stop_event);
+		if (stop && !pred->stop_event.after) {
+			debug("Stopped processing event log before indicated event\n");
+			break;
+		}
+
+		if (ev->pcr_index == pcr_index) {
+			/* Advance the event in the comparison event log */
+			while(ev_cmp) {
+				if (ev_cmp->pcr_index == pcr_index)
+					break;
+				ev_cmp = ev_cmp->next;
+			}
+
+			if (ev_cmp == NULL) {
+				tpm_event_print(ev);
+				printf("No corresponding event in the comparison event log\n");
+				printf("\n");
+				return false;
+			}
+
+			if (!(cmp_digest = tpm_event_get_digest(ev_cmp, pred->algo_info)))
+				fatal("Comparison event log lacks a hash for digest algorithm %s\n", pred->algo);
+			if (ev->predicted_digest.algo) {
+				predicted_digest = &ev->predicted_digest;
+
+				if (predicted_digest->size == cmp_digest->size
+				    && memcmp(predicted_digest->data, cmp_digest->data, cmp_digest->size)) {
+					printf("Predicted event:\n");
+					tpm_predicted_event_print(ev);
+					printf("Actual event:\n");
+					tpm_event_print(ev_cmp);
+					printf("\n");
+
+					return false;
+				}
+			}
+		}
+
+		if (stop) {
+			debug("Stopped processing event log after indicated event\n");
+			break;
+		}
+	}
+
+	return true;
+}
+
+static unsigned int
+predictor_compare(struct predictor *pred, struct predictor *pred_cmp)
+{
+	const tpm_pcr_bank_t *bank = &pred->prediction;
+	unsigned int pcr_index;
+	tpm_event_t *stop_event = NULL;
+	unsigned int num_diff = 0;
+
+	predictor_pre_scan_eventlog(pred, &stop_event);
+
+	for (pcr_index = 0; pcr_index < PCR_BANK_REGISTER_MAX; ++pcr_index) {
+		if (!pcr_bank_register_is_valid(bank, pcr_index))
+			continue;
+
+		if (!compare_events(pred, pred_cmp, pcr_index, stop_event))
+			num_diff++;
+	}
+
+	if (num_diff == 0)
+		printf("Predicted event log matches.\n");
+
+	return 0;
+}
+
 static void
 predictor_report(struct predictor *pred)
 {
@@ -1015,6 +1151,7 @@ int
 main(int argc, char **argv)
 {
 	struct predictor *pred;
+	struct predictor *pred_cmp;
 	int action = ACTION_NONE;
 	tpm_pcr_selection_t *pcr_selection = NULL;
 	char *opt_from = NULL;
@@ -1037,11 +1174,14 @@ main(int argc, char **argv)
 	char *opt_policy_name = NULL;
 	char *opt_target_platform = NULL;
 	char *opt_boot_entry = NULL;
+	bool opt_compare_current = false;
 	const target_platform_t *target;
 	unsigned int action_flags = 0;
 	unsigned int rsa_bits = 2048;
 	int c, exit_code = 0;
 
+	set_srk_alg("RSA");
+
 	while ((c = getopt_long(argc, argv, "dhA:CF:LSZ", options, NULL)) != EOF) {
 		switch (c) {
 		case 'A':
@@ -1109,6 +1249,9 @@ main(int argc, char **argv)
 		case OPT_RSA_BITS:
 			opt_rsa_bits = optarg;
 			break;
+		case OPT_ECC_SRK:
+			set_srk_alg("ECC");
+			break;
 		case OPT_INPUT:
 			opt_input = optarg;
 			break;
@@ -1135,6 +1278,9 @@ main(int argc, char **argv)
 		case OPT_TARGET_PLATFORM:
 			opt_target_platform = optarg;
 			break;
+		case OPT_COMPARE_CURRENT:
+			opt_compare_current = true;
+			break;
 		case 'h':
 			usage(0, NULL);
 		default:
@@ -1153,6 +1299,9 @@ main(int argc, char **argv)
 	if (opt_create_testcase)
 		runtime_record_testcase(testcase_alloc(opt_create_testcase));
 
+	if (!opt_replay_testcase && opt_compare_current)
+		fatal("--compare-current is only valid for --replay-testcase\n");
+
 	if (opt_rsa_bits) {
 		if (strcmp(opt_rsa_bits, "2048") == 0)
 			rsa_bits = 2048;
@@ -1320,12 +1469,24 @@ main(int argc, char **argv)
 	if (opt_stop_event)
 		predictor_set_stop_event(pred, opt_stop_event, !opt_stop_before);
 
+	if (opt_compare_current) {
+		testcase_t *tc_playback = runtime_get_replay_testcase();
+		/* Disable replay testcase temporarily to access the current TPM event log*/
+		runtime_replay_testcase(NULL);
+		pred_cmp = predictor_new(pcr_selection, "eventlog", NULL,
+					 opt_output_format, opt_boot_entry);
+		/* Restore replay testcase */
+		runtime_replay_testcase(tc_playback);
+	}
+
 	if (!predictor_update_all(pred, argc - optind, argv + optind))
 		return 1;
 
 	if (action == ACTION_PREDICT) {
 		if (opt_verify)
 			exit_code = !!predictor_verify(pred, opt_verify);
+		else if (opt_compare_current)
+			exit_code = !!predictor_compare(pred, pred_cmp);
 		else
 			predictor_report(pred);
 	} else
