@@ -31,6 +31,9 @@ extern uint32_t		esys_tr_rh_owner;
 extern ESYS_CONTEXT *	tss_esys_context(void);
 extern void		tss_print_error(int rc, const char *msg);
 
+extern TPM2B_PUBLIC *	tss_read_public_key(const char *);
+extern bool		tss_write_public_key(const char *, const TPM2B_PUBLIC *);
+
 static inline bool
 tss_check_error(int rc, const char *msg)
 {
