@@ -33,6 +33,9 @@ typedef struct cert_table	cert_table_t;
 typedef struct parsed_cert	parsed_cert_t;
 typedef struct pecoff_image_info pecoff_image_info_t;
 typedef struct testcase		testcase_t;
+typedef struct stored_key	stored_key_t;
+typedef struct target_platform	target_platform_t;
+typedef struct uapi_boot_entry	uapi_boot_entry_t;
 
 #endif /* TYPES_H */
 
